import '../agent/http/http_client.dart';
import 'action/interface/action.dart';
import 'action/null_web_request_timing.dart';
import 'interface/web_request_timing.dart';
import 'action/web_request_timing.dart';
import 'model/configuration.dart';
import 'interface/agent_interface.dart';
import 'action/null_actions/null_root_action_impl.dart';
import 'model/data_collection_level.dart';
import 'model/log_level.dart';
import 'model/user_privacy_options.dart';
import 'util/logger.dart';
import 'util/string_utils.dart';
import 'action/action.dart';
import 'action/counter.dart';
import 'action/interface/root_action.dart';
import 'model/platform.dart';

import 'dart:async';
import 'package:flutter/widgets.dart';
import 'package:flutter/services.dart';
import 'package:http/http.dart';

/// Main entry point for creating actions and events via Dynatrace
class DynatraceImpl implements Dynatrace {
  static Dynatrace? _instance;
  final MethodChannel _methodChannel;

  final Function _startupUpFunction;

  bool _started = false;
  bool _crashReporting = true;

  /// Factory method initializing singelton if not already created
  factory DynatraceImpl() {
    if (_instance == null) {
      const MethodChannel _methodChannel =
          MethodChannel('dynatrace_flutter_plugin/dynatrace');
      DynatraceImpl.private(_methodChannel);
    }

    return _instance as DynatraceImpl;
  }

  /// Private constructor for intializing main class
  DynatraceImpl.private(this._methodChannel,
      {Function startUpFunction = runApp})
      : _startupUpFunction = startUpFunction {
    _instance = this;
  }

  /// Returns if the plugin is already enabled or not. This depends if
  /// there is a configuration available which will only be set in the start
  /// method.
  bool _isEnabled() {
    return _started;
  }

  @override
  Future<void> start(Widget topLevelWidget,
      {Configuration configuration = const Configuration()}) async {
    // If our plugin fails to intialize, allow the app to run like normal
    await runZonedGuarded<Future<void>>(() async {
      await setConfig(configuration: configuration);
      _startupUpFunction(topLevelWidget);
    }, Dynatrace().reportZoneStacktrace);
  }

  @override
  Future<void> startWithoutWidget(
      {Configuration configuration = const Configuration()}) {
    return setConfig(configuration: configuration);
  }

  /// Internal helper method which sets the configuration
  Future<void> setConfig(
      {bool crashReport = true,
      Configuration configuration = const Configuration()}) async {
    try {
      WidgetsFlutterBinding.ensureInitialized();
      Map<String, bool>? autoStartConfiguration =
          await _methodChannel.invokeMapMethod("getAutoStartConfiguration");

      if (!_isEnabled()) {
        _started = true;

        if (autoStartConfiguration!["autoStart"]!) {
          // Take over the log level of native agent
          configuration = Configuration(
              logLevel: autoStartConfiguration["logLevel"]!
                  ? LogLevel.Debug
                  : LogLevel.Info,
              reportCrash: autoStartConfiguration["crashReporting"]!);
        } else {
          if (configuration.isAutoStartPropertyAvailable) {
            // We need to do a manual startup if it is available
            _methodChannel.invokeMethod(
                "start", configuration.getStartupConfiguration());
          }
        }

        Logger().logLevel = configuration.logLevel;

        _crashReporting = configuration.isCrashReportingEnabled;

        if (!configuration.isCrashReportingEnabled) {
          Logger().d("Crash reporting deactivated.", logType: LogType.Info);
        } else {
          Logger().d(
              "Dynatrace enableFlutterErrorCapturing(): Activating Crash reporting.",
              logType: LogType.Info);
          // Overrides FlutterError.onError to capture unhandled exceptions and report them
          enableFlutterErrorCapturing();
        }
      }

      Logger().i("Dynatrace Flutter Plugin started!", logType: LogType.Info);
    } catch (e) {
      Logger().i(
          "Dynatrace Flutter Plugin failed to initialize. Error: ${e.toString()}",
          logType: LogType.Error);
      _crashReporting = false;
      _started = false;
    }
  }

  /// Internal helper method which enables the error caputring within Flutter
  /// and overrides the public onError handler
  void enableFlutterErrorCapturing() {
    FlutterError.onError = (FlutterErrorDetails details) async {
      FlutterError.dumpErrorToConsole(details, forceReport: true);
      if (_crashReporting) {
        return reportErrorStacktrace(
            details.exception.runtimeType.toString() +
                ": " +
                details.summary.toString(),
            details.exception.runtimeType.toString(),
            details.summary.toString(),
            details.stack.toString());
      }
    };
  }

  @override
  DynatraceRootAction enterAction(String name, {Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace enterAction($name): Dynatrace().start() method was not called! Action will have no effect.",
          logType: LogType.Warning);
      return DynatraceNullRootAction();
    }

    if (StringUtils.isStringNullOrEmpty(name)) {
      Logger().d(
          "Dynatrace enterAction($name): Action Name was empty or null! Action will have no effect.",
          logType: LogType.Error);
      return DynatraceNullRootAction();
    }

    Logger().d("Dynatrace enterAction($name)", logType: LogType.Info);
    int key = ActionCounter().getCount();

    DynatraceRootAction action =
        DynatraceRootActionImpl.private(key, _methodChannel);
    ActionCounter().setOpenAction(key, action);

    _methodChannel.invokeMethod("enterAction", <String, dynamic>{
      "name": name,
      "key": key,
      "platform": platform?.index
    });

    return action;
  }

  /// Returns a web request tag for a certain [url]
  Future<String?> getHTTPTagForWebRequest(String url) async {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace getHTTPTagForWebRequest($url): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return "";
    }
    Logger()
        .d("Dynatrace getHTTPTagForWebRequest($url)", logType: LogType.Info);
    return await _methodChannel.invokeMethod(
        "getRequestTagForInterceptor", <String, dynamic>{"url": url});
  }

  /// Creates a web request timing event by using a tag as [header] and a [url]
  WebRequestTiming createWebRequestTiming(String header, String url) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace createWebRequestTiming($header, $url): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return NullWebRequestTiming();
    }
    Logger().d("Dynatrace createWebRequestTiming($header, $url)",
        logType: LogType.Info);
    return DynatraceWebRequestTiming(header, url, _methodChannel);
  }

  @override
  void reportStringValue(String? valueName, String? value,
      {Platform? platform}) {
    if (ActionCounter().getCurrentAction().runtimeType !=
        DynatraceNullRootAction) {
      Logger().d("Dynatrace reportStringValue($valueName, $value)",
          logType: LogType.Info);
      ActionCounter().getCurrentAction().reportStringValue(valueName, value);
    } else {
      Logger().d(
          "Dynatrace reportStringValue($valueName, $value): There are no open actions! String cannot be reported!",
          logType: LogType.Error);
    }
  }

  @override
  void reportIntValue(String? valueName, int? value, {Platform? platform}) {
    if (ActionCounter().getCurrentAction().runtimeType !=
        DynatraceNullRootAction) {
      Logger().d("Dynatrace reportIntValue($valueName, $value)",
          logType: LogType.Info);
      ActionCounter().getCurrentAction().reportIntValue(valueName, value);
    } else {
      Logger().d(
          "Dynatrace reportIntValue($valueName, $value): There are no open actions! Int cannot be reported!",
          logType: LogType.Error);
    }
  }

  @override
  void reportDoubleValue(String? valueName, double? value,
      {Platform? platform}) {
    if (ActionCounter().getCurrentAction().runtimeType !=
        DynatraceNullRootAction) {
      Logger().d("Dynatrace reportDoubleValue($valueName, $value)",
          logType: LogType.Info);
      ActionCounter().getCurrentAction().reportDoubleValue(valueName, value);
    } else {
      Logger().d(
          "Dynatrace reportDoubleValue($valueName, $value): There are no open actions! Double cannot be reported!",
          logType: LogType.Error);
    }
  }

  @override
  void reportEvent(String? eventName, {Platform? platform}) {
    if (ActionCounter().getCurrentAction().runtimeType !=
        DynatraceNullRootAction) {
      Logger().d("Dynatrace reportEvent($eventName)", logType: LogType.Info);
      ActionCounter().getCurrentAction().reportEvent(eventName);
    } else {
      Logger().d(
          "Dynatrace reportEvent($eventName): There are no open actions! Event cannot be reported!",
          logType: LogType.Error);
    }
  }

  @override
  void reportErrorInAction(String? errorName, int? errorCode,
      {Platform? platform}) {
    if (ActionCounter().getCurrentAction().runtimeType !=
        DynatraceNullRootAction) {
      Logger().d("Dynatrace reportErrorInAction($errorName, $errorCode)",
          logType: LogType.Info);
      ActionCounter().getCurrentAction().reportError(errorName, errorCode);
    } else {
      Logger().d(
          "Dynatrace reportErrorInAction($errorName, $errorCode): There are no open actions! Error cannot be reported!",
          logType: LogType.Error);
    }
  }

  @override
  void reportError(String? errorName, int? errorCode, {Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace reportError($errorName, $errorCode): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    if (!StringUtils.isStringNullOrEmpty(errorName) && errorCode != null) {
      Logger().d("Dynatrace reportError($errorName, $errorCode)",
          logType: LogType.Info);
      _methodChannel.invokeMethod("reportError", <String, dynamic>{
        "errorName": errorName,
        "errorCode": errorCode,
        "platform": platform?.index
      });
    } else {
      Logger().d(
          "Dynatrace reportError(errorName, errorCode): Error was not reported because errorName is null or empty and or errorCode is null! \nerrorName: $errorName \nerrorCode: $errorCode",
          logType: LogType.Error);
    }
  }

  @override
  void reportErrorStacktrace(
      String errorName, String errorValue, String reason, String stacktrace,
      {Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace reportErrorStacktrace($errorName, $errorValue, $reason, $stacktrace): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    if (!StringUtils.isStringNullOrEmpty(errorName)) {
      Logger().d(
          "Dynatrace reportErrorStacktrace($errorName, $errorValue, $reason, $stacktrace)",
          logType: LogType.Info);
      _methodChannel.invokeMethod("reportErrorStacktrace", <String, dynamic>{
        "errorName": errorName,
        "errorValue": errorValue,
        "reason": reason,
        "stacktrace": stacktrace,
        "platform": platform?.index
      });
    } else {
      Logger().d(
          "Dynatrace reportErrorStacktrace(errorName, errorValue, reason, stacktrace): Error will not be reported because errorName is null or empty! \nerrorName: $errorName \nerrorValue: $errorValue \nreason: $reason \nstacktrace: $stacktrace",
          logType: LogType.Error);
    }
  }

  @override
  Future<void> reportZoneStacktrace(dynamic error, StackTrace stacktrace,
      {Platform? platform}) async {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace reportZoneStacktrace(error, stacktrace): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }

    String runtimeType = error.runtimeType.toString();

    String message = "";

    try {
      if (error.message != null) {
        message = error.message.toString();
      }
    } catch (e) {
      // Nothing happens on purpose
    }

    if (_crashReporting &&
        error != null &&
        !StringUtils.isStringNullOrEmpty(runtimeType)) {
      Logger().d(
          "Dynatrace reportZoneStacktrace($runtimeType: $message, $stacktrace)",
          logType: LogType.Info);
      await _methodChannel
          .invokeMethod("reportErrorStacktrace", <String, dynamic>{
        "errorName": "$runtimeType: $message",
        "errorValue": runtimeType,
        "reason": message,
        "stacktrace": stacktrace.toString(),
        "platform": platform?.index
      });
    } else {
      Logger().d(
          "Dynatrace reportZoneStacktrace(error, stacktrace): Error will not be reported because something is missing! \nerrorName: $runtimeType: $message \nstacktrace: ${stacktrace.toString()}",
          logType: LogType.Error);
    }
  }

  @override
  Future<void> reportCrashWithException(
      String crashName, Exception exceptionObject,
      {String? reason = "-", Platform? platform}) async {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace reportCrashWithException($crashName, ${exceptionObject.toString()}): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
    }

    if (!StringUtils.isStringNullOrEmpty(crashName) &&
        !StringUtils.isStringNullOrEmpty(exceptionObject.toString())) {
      if (_crashReporting) {
        Logger().d(
            "Dynatrace reportCrashWithException($crashName, ${exceptionObject.toString()})",
            logType: LogType.Info);
        await _methodChannel
            .invokeMethod("reportCrashWithException", <String, dynamic>{
          "crashName": crashName,
          "reason": reason,
          "stacktrace": exceptionObject.toString(),
          "platform": platform?.index
        });
      }
    } else {
      Logger().d(
          "Dynatrace reportCrashWithException(crashName, exceptionObject): Crash will not be reported because crashName/exceptionObject is null or empty! \ncrashName: $crashName \nexceptionObject: ${exceptionObject.toString()} \nreason:${reason != "-" ? reason : "named/optional parameter not used"}",
          logType: LogType.Error);
    }
  }

  @override
  Future<void> reportCrash(String? errorName, String reason, String stacktrace,
      {Platform? platform}) async {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace reportCrash($errorName, $reason, $stacktrace): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
    }

    if (!StringUtils.isStringNullOrEmpty(errorName)) {
      if (_crashReporting) {
        Logger().d("Dynatrace reportCrash($errorName, $reason, $stacktrace)",
            logType: LogType.Info);
        await _methodChannel.invokeMethod("reportCrash", <String, dynamic>{
          "errorValue": errorName,
          "reason": reason,
          "stacktrace": stacktrace,
          "platform": platform?.index
        });
      }
    } else {
      Logger().d(
          "Dynatrace reportCrash(errorName, reason, stacktrace): Crash will not be reported because errorName is null or empty! \nerrorName: $errorName \nreason: $reason \nstacktrace: $stacktrace",
          logType: LogType.Error);
    }
  }

  void sendEvent(String name, Map<String, dynamic>? attributes,
      {Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace sendEvent($name, ${attributes.toString()}): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
    } else {
      if (StringUtils.isStringNullOrEmpty(name)) {
        Logger().d(
            "Dynatrace sendEvent($name, ${attributes.toString()}): Name must be a non empty string!",
            logType: LogType.Error);
        return;
      }

      attributes ??= {};

      //var internalAttributes = {'dt.agent.flavor': 'flutter', ...attributes};
      var internalAttributes = {...attributes};

      Logger().d("Dynatrace sendEvent($name, ${internalAttributes.toString()})",
          logType: LogType.Info);

      _methodChannel.invokeMethod("sendEvent", <String, dynamic>{
        "name": name,
        "attributes": internalAttributes,
        "platform": platform?.index
      });
    }
  }

  @override
  void sendBizEvent(String type, Map<String, dynamic>? attributes,
      {Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace sendBizEvent($type, ${attributes.toString()}): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
    } else {
      if (StringUtils.isStringNullOrEmpty(type)) {
        Logger().d(
            "Dynatrace sendBizEvent($type, ${attributes.toString()}): Type must be a non empty string!",
            logType: LogType.Error);
        return;
      }

      attributes ??= {};

      //var internalAttributes = {'dt.agent.flavor': 'flutter', ...attributes};
      var internalAttributes = {...attributes};

      Logger().d(
          "Dynatrace sendBizEvent($type, ${internalAttributes.toString()})",
          logType: LogType.Info);

      _methodChannel.invokeMethod("sendBizEvent", <String, dynamic>{
        "type": type,
        "attributes": internalAttributes,
        "platform": platform?.index
      });
    }
  }

  @override
  void endSession({Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace endSession(): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
    } else {
      Logger().d("Dynatrace endSession()", logType: LogType.Info);
      _methodChannel.invokeMethod(
          "endVisit", <String, dynamic>{"platform": platform?.index});
    }
  }

  @override
  void identifyUser(String? user, {Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace identifyUser($user): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }
    Logger().d("Dynatrace identifyUser($user)", logType: LogType.Info);
    _methodChannel.invokeMethod("identifyUser",
        <String, dynamic>{"user": user, "platform": platform?.index});
  }

  @override
  void setGPSLocation(double latitude, double longitude, {Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace setGPSLocation($latitude, $longitude): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return;
    }
    Logger().d("Dynatrace setGPSLocation($latitude, $longitude)",
        logType: LogType.Info);
    _methodChannel.invokeMethod("setGPSLocation", <String, dynamic>{
      "latitude": latitude,
      "longitude": longitude,
      "platform": platform?.index
    });
  }

  @override
  void flushEvents({Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace flushEvents(): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
    } else {
      Logger().d("Dynatrace flushEvents()", logType: LogType.Info);
      _methodChannel.invokeMethod(
          "flushEvents", <String, dynamic>{"platform": platform?.index});
    }
  }

  @override
  void applyUserPrivacyOptions(UserPrivacyOptions userPrivacyOptions,
      {Platform? platform}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace applyUserPrivacyOptions(userPrivacyOptions): Dynatrace().start() method was not called! User Privacy Options can't be applied!",
          logType: LogType.Warning);
    } else {
      Logger().d(
          "Dynatrace applyUserPrivacyOptions({ ${userPrivacyOptions.crashReportingOptedIn}, ${userPrivacyOptions.dataCollectionLevel.index == 3 ? 2 : userPrivacyOptions.dataCollectionLevel.index} })",
          logType: LogType.Info);
      _methodChannel.invokeMethod("applyUserPrivacyOptions", <String, dynamic>{
        "crashReportingOptedIn": userPrivacyOptions.crashReportingOptedIn,
        "dataCollectionLevel": userPrivacyOptions.dataCollectionLevel.index == 3
            ? 2
            : userPrivacyOptions.dataCollectionLevel.index,
        "platform": platform?.index
      });
    }
  }

  @override
  Future<UserPrivacyOptions> getUserPrivacyOptions({Platform? platform}) async {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace getUserPrivacyOptions(): Dynatrace().start() method was not called! User Privacy Options are not available!",
          logType: LogType.Warning);
      return UserPrivacyOptions(DataCollectionLevel.Off, false);
    } else {
      Logger().d("Dynatrace getUserPrivacyOptions()", logType: LogType.Info);
      Map<String, dynamic>? userPrivacyOptions = await (_methodChannel
          .invokeMapMethod("getUserPrivacyOptions",
              <String, dynamic>{"platform": platform?.index}));

      if (userPrivacyOptions != null) {
        Logger().d('''Dynatrace getUserPrivacyOptions(): Successful! 
          \nDataCollectionLevel: ${DataCollectionLevel.values[userPrivacyOptions["dataCollectionLevel"]]},
          \nCrashReportingOptedIn: ${userPrivacyOptions["crashReportingOptedIn"]}''',
            logType: LogType.Info);
        return UserPrivacyOptions(
            DataCollectionLevel
                .values[userPrivacyOptions["dataCollectionLevel"]],
            userPrivacyOptions["crashReportingOptedIn"]);
      } else {
        Logger().d(
            "Dynatrace getUserPrivacyOptions(): User Privacy Options called failed! Returning default values!",
            logType: LogType.Warning);
        return UserPrivacyOptions(DataCollectionLevel.Off, false);
      }
    }
  }

  @override
  DynatraceHttpClient createHttpClient({Client? client}) {
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace createHttpClient(): Dynatrace().start() method was not called! DynatraceHttpClient is not available!",
          logType: LogType.Warning);
    } else {
      Logger().d("Dynatrace createHttpClient()", logType: LogType.Info);
    }
    return client != null
        ? DynatraceHttpClient(_instance as DynatraceImpl, client: client)
        : DynatraceHttpClient(_instance as DynatraceImpl);
  }

  @override
  DynatraceAction getCurrentAction() {
    var action = ActionCounter().getCurrentAction();
    if (!_isEnabled()) {
      Logger().i(
          "Dynatrace getCurentAction(): Dynatrace().start() method was not called!",
          logType: LogType.Warning);
      return DynatraceNullRootAction();
    }

    if (action.runtimeType != DynatraceNullRootAction) {
      Logger().d("Dynatrace getCurrentAction(): ${action.runtimeType}",
          logType: LogType.Info);
    } else {
      Logger().d("Dynatrace getCurrentAction(): There are no open actions!",
          logType: LogType.Error);
    }

    return action;
  }
}
