import '../../agent/util/logger.dart';
import '../../cli/config/model/cli_arguments.dart';
import '../../cli/config/model/ios_config.dart';
import '../../cli/util/util.dart';

import 'dart:io';

/// Responsible for reading the configuration string which contains
/// all the plist values for the iOS Agent
class IosConfigurator {
  late File _pListFile;
  late Logger _logger;
  late Util _util;

  /// Constructor which uses the [arguments] passed to the CLI and
  /// [pListFile], [util] and [logger] which will typically be used for tests
  IosConfigurator(CommandLineArguments arguments, File pListFile, Util util,
      Logger logger) {
    _pListFile = pListFile;
    _util = util;
    _logger = logger;
  }

  /// Loads the [pathToPlist] and checks if it is really a plist file. It removes the Dynatrace configuration
  /// and inserts the new configuration stored in [iosConfig].
  Future<void> modifyPListFile(IosConfiguration iosConfig,
      {bool? isUninstall = false}) async {
    _logger.d("Current operating system: ${_util.currentOS()}",
        logType: LogType.Info);
    if (_util.isMacOS()) {
      if (!_pListFile.path.endsWith(".plist") || !await _pListFile.exists()) {
        throw FileSystemException(
            "Can't find .plist file. plist path must also include the plist file! : ${_pListFile.path}");
      }

      String tempPlistPath =
          _pListFile.path.replaceAll("Info.plist", "Dynatrace.plist");

      String pListContent = await _pListFile.readAsString();

      if (iosConfig.getConfig() == null || iosConfig.getConfig()!.isEmpty) {
        _logger.i(
            "Can't write configuration of iOS agent because it is missing!",
            logType: LogType.Warning);
      } else {
        if (isUninstall!) {
          _logger.d("Removing Dynatrace properties from ${_pListFile.path}!",
              logType: LogType.Info);
        }

        await removePropertiesFromPlist(
            getDynatracePropsFromPlist(pListContent), _pListFile.path);

        if (!isUninstall) {
          createTempPlist(iosConfig);
          // Combine/Add dynatrace config to Info.plist
          await mergePlists(tempPlistPath, _pListFile.path);
        }
      }
    } else {
      if (isUninstall!) {
        _logger.i(
            "Unable to remove Dynatrace properties from ${_pListFile.path} as current operating system is not MacOS!",
            logType: LogType.Warning);
      } else {
        _logger.i(
            "Can't write configuration of iOS agent as current operating system is not MacOS!",
            logType: LogType.Warning);
      }
    }
  }

  /// Creates a temporary plist and inserts the new configuration stored in [iosConfig].
  File createTempPlist(IosConfiguration iosConfig) {
    String tempPlistStr = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
  <dict>
    ${iosConfig.getConfig()}
    <key>DTXFlavor</key>
    <string>flutter</string>
  </dict>
</plist>""";

    File tempPlist =
        File(_pListFile.path.replaceAll("Info.plist", "Dynatrace.plist"));
    try {
      tempPlist.writeAsStringSync(tempPlistStr);
    } catch (e) {
      _logger.i("Unable to create temporary plist! Error: $e",
          logType: LogType.Warning);
    }

    return tempPlist;
  }

  /// Deletes the temporary plist used to add the [iosConfig] to the Info.plist
  deleteTempPlist(File plist) {
    try {
      plist.deleteSync();
    } catch (e) {
      _logger.i("Unable to delete temporary .plist. Error: $e",
          logType: LogType.Warning);
    }
    if (!plist.existsSync()) {
      _logger.d("Successfully deleted temporary .plist", logType: LogType.Info);
    }
  }

  /// Combines the temporary plist with the dynatrace properties and the Info.plist
  Future<void> mergePlists(String tempPlistPath, String infoPlistPath) async {
    try {
      await Process.run('/usr/libexec/plistbuddy',
          ['-x', '-c', 'Merge $tempPlistPath', infoPlistPath]);
    } catch (e) {
      _logger.i(
          "Unable to update Info.plist with dynatrace properties. Error: $e",
          logType: LogType.Warning);
    }
  }

  /// Deletes dynatrace properties ([props]) from the Info.plist during uninstall
  Future<void> removePropertiesFromPlist(
      List<String> props, String infoPlistPath) async {
    try {
      List<String> commands = ['-x'];

      for (var i = 0; i < props.length; i++) {
        commands.add('-c');
        commands.add('Delete :${props[i]}');
      }
      commands.add(infoPlistPath);

      await Process.run('/usr/libexec/plistbuddy', commands);
    } catch (e) {
      _logger.i(
          "Can't remove configuration of iOS agent from $infoPlistPath! Error: $e",
          logType: LogType.Error);
    }
  }

  /// Parse the plist [content] and return a list of dynatrace properties
  List<String> getDynatracePropsFromPlist(String content) {
    List<String> props = content.split("<key>");
    props.removeWhere((item) => (!item.contains("DTX")));

    for (var i = 0; i < props.length; i++) {
      props[i] = props[i]
          .substring(props[i].indexOf("DTX"), props[i].indexOf("</key>"));
    }
    return props;
  }
}
